"""txcelery
Copyright Sentimens Research Group, LLC
2014
MIT License

Module Contents:
    - DeferredTask
    - CeleryClient
"""
import logging

import redis
from celery.exceptions import TimeoutError
from twisted.internet import defer, reactor
from twisted.internet.defer import inlineCallbacks, DeferredSemaphore
from twisted.internet.threads import deferToThread
from twisted.python.failure import Failure

logger = logging.getLogger(__name__)


class _DeferredTaskPatch(defer.Deferred):
    """Subclass of `twisted.defer.Deferred` that wraps a
    `celery.local.PromiseProxy` (i.e. a "Celery task"), exposing the combined
    functionality of both classes.

    `_DeferredTask` instances can be treated both like ordinary Deferreds and
    oridnary PromiseProxies.
    """

    #: Wait Period
    MAX_RETRIES = 3

    __reactorShuttingDown = False
    __dbSessionCreator = None
    __sqlaUrl = None
    __deferredSemaphore: DeferredSemaphore = None

    @classmethod
    def setupPostGreSQLConnection(cls, dbSessionCreator,
                                  sqlaUrl: str,
                                  parallelism: int):
        """ Start Celery Threads

        Configure the Celery connection settings.
        :param dbSessionCreator: A callable that will fiv.

        :param sqlaUrl: The connection string the SQLA Engine in PostGreSQL will use.

        :param parallelism: The number of tasks to run in parallel at one time

        """
        _DeferredTaskPatch.__dbSessionCreator = dbSessionCreator
        _DeferredTaskPatch.__sqlaUrl = sqlaUrl
        _DeferredTaskPatch.__deferredSemaphore = DeferredSemaphore(parallelism)

        from txcelery import defer
        defer._DeferredTask = _DeferredTaskPatch

        reactor.addSystemEventTrigger(
            "before", "shutdown", cls.setReactorShuttingDown
        )

    @classmethod
    def setReactorShuttingDown(cls):
        cls.__reactorShuttingDown = True

    def __init__(self, func, *args, **kwargs):
        """Instantiate a `_DeferredTask`.  See `help(_DeferredTask)` for details
        pertaining to functionality.

        :param async_result : celery.result.AsyncResult
            AsyncResult to be monitored.  When completed or failed, the
            _DeferredTask will callback or errback, respectively.
        """
        # Deferred is an old-style class
        defer.Deferred.__init__(self, self._canceller)
        self.addErrback(self._cbErrback)

        # Auto initialise the the thread pool if the app hasn't already done so

        self.__retries = self.MAX_RETRIES

        d = self._start(func, *args, **kwargs)
        d.addBoth(self._threadFinishInMain)

    @inlineCallbacks
    def _start(self, func, *args, **kwargs):
        while self.__retries and not self.called and not self.__reactorShuttingDown:
            self.__retries -= 1
            try:
                result = yield self.__deferredSemaphore \
                    .run(deferToThread, self._run, func, *args, **kwargs)
                return result

            except Exception as e:
                if not self.__retries:
                    print(e.__class__.__name__)
                    print(e)
                    raise

    def addTimeout(self, timeout, clock, onTimeoutCancel=None):
        defer.Deferred.addTimeout(self, timeout, clock, onTimeoutCancel=onTimeoutCancel)

    def _threadFinishInMain(self, result):
        if self.called:
            return

        if isinstance(result, Failure):
            if result.check(redis.exceptions.ConnectionError) and self.__retries:
                self.__retries -= 1

            self.errback(result)

        else:
            self.callback(result)

    def _canceller(self, *args):
        self.__retries = 0

    def _cbErrback(self, failure: Failure) -> Failure:
        if isinstance(failure.value, TimeoutError):
            self._canceller()

        return failure

    def _run(self, func, *args, **kwargs):
        """ Monitor Task In Thread

        The Celery task state must be checked in a thread, otherwise it blocks.

        This may stuff with Celerys connection to the result backend.
        I'm not sure how it manages that.

        """
        from peek_storage.plpython.RunWorkerTaskPyInPg import runPyWorkerTaskInPgBlocking

        return runPyWorkerTaskInPgBlocking(self.__dbSessionCreator,
                                           self.__sqlaUrl,
                                           func,
                                           *args,
                                           **kwargs)



import logging

import os
from abc import ABCMeta, abstractmethod

logger = logging.getLogger(__name__)


class BuilderABC(metaclass=ABCMeta):

    def _writeFileIfRequired(self, dir, fileName, contents):
        fullFilePath = os.path.join(dir, fileName)

        # Apply any changes to these files using the transform code
        contents = self._syncFileHook(fileName, contents.encode()).decode()

        # Since writing the file again changes the date/time,
        # this messes with the self._recompileRequiredCheck
        if os.path.isfile(fullFilePath):
            with open(fullFilePath, 'r') as f:
                if contents == f.read():
                    logger.debug("%s is up to date", fileName)
                    return

        logger.debug("Writing new %s", fileName)

        with open(fullFilePath, 'w') as f:
            f.write(contents)


    @abstractmethod
    def _syncFileHook(self, fileName: str, contents: bytes) -> bytes:
        """ Sync File Hook

        see FrontendFileSync._syncFileHook

        """
        pass

    def _recompileRequiredCheck(self, feBuildDir: str, hashFileName: str) -> bool:
        """ Recompile Check

        This command lists the details of the source dir to see if a recompile is needed

        The find command outputs the following

        543101    0 -rw-r--r--   1 peek     sudo            0 Nov 29 17:27 ./src/app/environment/environment.component.css
        543403    4 drwxr-xr-x   2 peek     sudo         4096 Dec  2 17:37 ./src/app/environment/env-worker
        543446    4 -rw-r--r--   1 peek     sudo         1531 Dec  2 17:37 ./src/app/environment/env-worker/env-worker.component.html

        """

        excludeFilesEndWith = (".git", ".idea", '.lastHash')
        excludeFilesStartWith = ()

        def dirCheck(path):
            s = os.path.sep
            excludePathContains = ('__pycache__', 'node_modules', 'platforms', 'dist')

            # Always include the node_modules/@peek module dir
            if path.endswith(s + "@peek") or (s + "@peek" + s) in path:
                return True

            for exPath in excludePathContains:
                # EG "C:\thing\node_modules"
                if path.endswith(s + exPath):
                    return False

                # EG "C:\thing\node_modules\thing"
                if (s + exPath + s) in path:
                    return False

            return True

        fileList = []

        for (path, directories, filenames) in os.walk(feBuildDir):
            if not dirCheck(path):
                continue

            for filename in filenames:
                if [e for e in excludeFilesEndWith if filename.endswith(e)]:
                    continue

                if [e for e in excludeFilesStartWith if filename.startswith(e)]:
                    continue

                fullPath = os.path.join(path, filename)
                relPath = fullPath[len(feBuildDir) + 1:]
                stat = os.stat(fullPath)
                fileList.append('%s %s %s' % (relPath, stat.st_size, stat.st_mtime))

        newHash = '\n'.join(fileList)
        fileHash = ""

        if os.path.isfile(hashFileName):
            with open(hashFileName, 'r') as f:
                fileHash = f.read()

        fileHashLines = set(fileHash.splitlines())
        newHashLines = set(newHash.splitlines())
        changes = False

        for line in fileHashLines - newHashLines:
            changes = True
            logger.debug("Removed %s" % line)

        for line in newHashLines - fileHashLines:
            changes = True
            logger.debug("Added %s" % line)

        if changes:
            with open(hashFileName, 'w') as f:
                f.write(newHash)

        return changes



import logging
import os
import shutil
import unittest

import peek_platform
from jsoncfg.functions import config_to_json_str
from peek_platform.file_config.PeekFileConfigABC import PeekFileConfigABC
from peek_platform.file_config.PeekFileConfigPeekServerClientMixin import \
    PeekFileConfigPeekServerClientMixin
from peek_platform.file_config.PeekFileConfigPlatformMixin import \
    PeekFileConfigPlatformMixin

logging.basicConfig(level=logging.DEBUG)

logger = logging.getLogger(__name__)


class TestFileConfig(PeekFileConfigABC,
                     PeekFileConfigPeekServerClientMixin,
                     PeekFileConfigPlatformMixin):
    pass


class PeekFileConfigTest(unittest.TestCase):
    COMPONENT_NAME = "unit_test"
    HOME_DIR = os.path.expanduser('~/%s.home' % COMPONENT_NAME)
    CONFIG_FILE_PATH = os.path.join(HOME_DIR, 'config.json')

    def _rmHome(self):
        if os.path.exists(self.HOME_DIR):
            shutil.rmtree(self.HOME_DIR)

    def setUp(self):
        TestFileConfig._PeekFileConfigBase__instance = None
        peek_platform.PeekPlatformConfig.componentName = "unit_test"
        self._rmHome()
        os.makedirs(self.HOME_DIR, TestFileConfig.DEFAULT_DIR_CHMOD)

        with open(self.CONFIG_FILE_PATH, 'w') as fobj:
            fobj.write('{"nothing":{"is_true":true}}')

    def tearDown(self):
        self._rmHome()

    def testReadingConfig(self):
        bas = TestFileConfig()

        logger.info('nothing.is_true = %s', bas._cfg.nothing.is_true)
        self.assertTrue(bas._cfg.nothing.is_true())

    def testPeekServerDetails(self):
        bas = TestFileConfig()
        logger.info('peekServerPort = %s', bas.peekServerPort)
        logger.info('peekServerHost = %s', bas.peekServerHost)

        logger.info(config_to_json_str(bas._cfg))

        bas._save()

        with open(self.CONFIG_FILE_PATH, 'r') as fobj:
            logger.info(fobj.read())

    def testAsignment(self):
        inArr = [1, 2, 5, "20"]
        valuey = "valuey"
        twoFiveSix = 256

        bas = TestFileConfig()

        bas._cfg.op1.thingx = valuey
        bas._cfg.op1.thingy = twoFiveSix
        bas._cfg.op2.op3.arr = inArr

        logger.info(config_to_json_str(bas._cfg))

        bas._save()

        with open(self.CONFIG_FILE_PATH, 'r') as fobj:
            logger.info(fobj.read())

        TestFileConfig._PeekFileConfigBase__instance = None

        bas = TestFileConfig()
        self.assertEqual(bas._cfg.op2.op3.arr(), inArr)
        self.assertEqual(bas._cfg.op1.thingx(), valuey)
        self.assertEqual(bas._cfg.op1.thingy(), twoFiveSix)

    def testPlatformDetails(self):
        pluginName = 'plugin_noop'

        bas = TestFileConfig()

        # Defaults
        logger.info('plugin.plugin_noop.version = %s', bas.pluginVersion(pluginName))
        logger.info('platformVersion = %s', bas.platformVersion)

        bas.platformVersion = '4.4.4'
        bas.setPluginVersion(pluginName, '2.5.6')

        with open(self.CONFIG_FILE_PATH, 'r') as fobj:
            logger.info(fobj.read())

        TestFileConfig._PeekFileConfigBase__instance = None

        logger.info('plugin.plugin_noop.version = %s', bas.pluginVersion(pluginName))
        logger.info('platformVersion = %s', bas.platformVersion)

        self.assertEqual(bas.platformVersion, '4.4.4')
        self.assertEqual(bas.pluginVersion(pluginName), '2.5.6')



import logging
import os
import subprocess
from typing import Optional

from peek_platform.WindowsPatch import isWindows

logger = logging.getLogger(__name__)


class SpawnOsCommandException(subprocess.CalledProcessError):
    """ Spawn OS Command Exception

    This exception is raised when an OS command fails or returns a non-zero exit code

    """

    def __init__(self, returncode: int, cmd: str,
                 stdout: Optional[str] = None, stderr: Optional[str] = None,
                 message: Optional[str] = None):
        """ Constructor

        :param returncode: The code returned by the process.
        :param cmd: A string representing the command and args executed by bash.
        :param stdout: The output from the process.
        :param stderr: The error from the process.
        :param message: An additional message that can be used to emulate the message
        method of standard exceptions
        """

        subprocess.CalledProcessError.__init__(self, returncode, cmd, stdout, stderr)
        self.message = message

    def __str__(self):
        return ("%s\nCommand '%s' returned non-zero exit status %d"
                % (self.message, self.cmd, self.returncode))


class PtyOutParser:
    """ PTY Out Parser

    The node tools require a tty, so we run it with

        parser = _PtyOutParser()
        import pty
        pty.spawn(\*args, parser.read)

    The only problem being that the output is sent to stdout, to solve this we intercept
    the output, return a . (dot) for every read, which it sends to stdout, and then only log
    the summary at the end of the webpack build.

    """

    def __init__(self, loggingStartMarker=None):
        """ Constructor

        :param loggingStartMarker: If this is set, logging will not start until a line
        is starts with this marker.

        """
        self.data = ''
        self.startLogging = False  # Ignore all the stuff before the final summary
        self.allData = ''
        self.loggingStartMarker = loggingStartMarker

    def read(self, fd, size: Optional[int] = 1024):
        data = os.read(fd, size)
        self.data += data.decode(errors='ignore')
        self.allData += data.decode(errors='ignore')
        self.splitData()

        # Silence all the output
        if len(data):
            return b'.'

        # If there is no output, return the EOF data
        return data

    def splitData(self):
        lines = self.data.splitlines(True)
        lines.reverse()
        while lines:
            line = lines.pop()
            if not line.endswith(os.linesep):
                self.data = line
                break
            self.logData(line.strip(os.linesep))

    def logData(self, line):
        self.startLogging = (self.startLogging
                             or (self.loggingStartMarker
                                 and line.startswith(self.loggingStartMarker)))

        if not (line and self.startLogging):
            return

        logger.debug(line)


def spawnPty(cmdAndArgs: str,
             parser: PtyOutParser = PtyOutParser()) -> None:
    """ Spawn PTY

    This function spawns a PTY in Linux and falls back to using subprocess on windows.

    The commands are run via bash on both windows and posix compliant OSs. GitBash is
    recommended for windows.

    :param cmdAndArgs: A string containing the command and arguments to pass to bash.

    :param parser: An instance of c{PtyOutParser} that intercepts the PTY output.

    :returns: None, This either succeeds or raises an exception.

    :raises: c{CalledProcessError}
    """

    # If this is a posix compliant environment, then use pty
    if not isWindows:
        __reallySpawnSubprocess(cmdAndArgs, parser)
        return

    # Else, use subprocess
    spawnSubprocess(cmdAndArgs)


def __reallySpawnSubprocess(cmdAndArgs: str,
                            parser: PtyOutParser = PtyOutParser()) -> None:
    from peek_platform import PeekPlatformConfig
    bashExec = PeekPlatformConfig.config.bashLocation

    logger.debug("Using interpreter : %s", bashExec)
    logger.debug("Running command via subprocess : %s", cmdAndArgs)

    import pty
    exitCode = pty.spawn([bashExec, "-l", "-c", cmdAndArgs], parser.read, parser.read)

    if exitCode:
        raise SpawnOsCommandException(
            exitCode,
            cmdAndArgs,
            stdout="",
            stderr=parser.allData)


def spawnSubprocess(cmdAndArgs: str,
                    parser: Optional[PtyOutParser] = None) -> None:
    """ Spawn Subprocess

    This method calls an OS command using the subprocess package and bash as the
    interpreter.

    :param cmdAndArgs: A string containing the command and arguments to pass to bash.
    :param parser: The parser for the command output.

    :returns: None, This either succeeds or raises an exception.

    :raises: c{CalledProcessError}

    """
    from peek_platform import PeekPlatformConfig
    bashExec = PeekPlatformConfig.config.bashLocation

    logger.debug("Using interpreter : %s", bashExec)
    logger.debug("Running command via subprocess : %s", cmdAndArgs)

    commandComplete = subprocess.run([cmdAndArgs],
                                     executable=bashExec,
                                     stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                     shell=True)

    if parser:
        parser.read(commandComplete.stdout, size=None)
        parser.read(commandComplete.stderr, size=None)

    if commandComplete.returncode:
        raise SpawnOsCommandException(
            commandComplete.returncode,
            cmdAndArgs,
            stdout=commandComplete.stdout.decode(),
            stderr=commandComplete.stderr.decode())

    return commandComplete


def logSpawnException(exception: Exception) -> None:
    """ Log Exception

    This method logs the stderr / stdout data from an exception to the logger.

    :param exception: The exception to log the data from.
    """

    logger.exception(exception)

    if hasattr(exception, "message"):
        logger.error(exception.message)

    if hasattr(exception, "stderr"):
        [logger.error(l) for l in exception.stderr.splitlines()]

    if hasattr(exception, "stderr"):
        [logger.error(l) for l in exception.stderr.splitlines()]
